from pydantic.v1 import (
    BaseModel,
    Field,
)

from mxcubeweb.core.models.configmodels import ModeEnum


class SimpleNameValue(BaseModel):
    name: str
    # It's important to have str before bool, to avoid issue with bool being casted to string
    value: bool | str | int


class AppSettingsModel(BaseModel):
    mode: ModeEnum = Field(ModeEnum.OSC, description="MXCuBE mode SSX or OSC")
    version: str = Field("", description="MXCuBE version")
    mesh_result_format: str = Field(
        "PNG", description="Format of mesh result for display"
    )
    use_native_mesh: bool = Field(
        True,
        description=(
            "Usage of native mesh feature, true by default. The native mesh feature can"
            "be dis-activated to not clash with i.e workflow mesh,"
        ),
    )

    enable_2d_points: bool = Field(
        True,
        description=(
            " Enable features to work with points in the plane, called2D-points,"
            " (none centred positions)"
        ),
    )

    class Config:
        extra: "forbid"  # noqa: F821



from gevent import monkey

monkey.patch_all(thread=False)

# Disabling E402 (module level import not at top of file)
# for the lines below as we are monkey patching
import argparse  # noqa: E402
import os  # noqa: E402
import sys  # noqa: E402
import traceback  # noqa: E402
from pathlib import Path  # noqa: E402
from unittest import mock  # noqa: E402

from mxcubecore import HardwareRepository as HWR  # noqa: E402

from mxcubeweb.app import MXCUBEApplication as mxcube  # noqa: E402
from mxcubeweb.config import Config  # noqa: E402
from mxcubeweb.server import Server as server  # noqa: E402

sys.modules["Qub"] = mock.Mock()
sys.modules["Qub.CTools"] = mock.Mock()


def parse_args(argv):
    # by default load the 'demo' mocked beamline
    hwr_directory = str(Path(Path(__file__).parents[1], "demo"))

    opt_parser = argparse.ArgumentParser(
        description="mxcube-web Backend server command line utility."
    )

    opt_parser.add_argument(
        "-r",
        "--repository",
        dest="hwr_directory",
        help="Hardware Repository XML files path",
        default=hwr_directory,
    )

    opt_parser.add_argument(
        "-s",
        "--static-folder",
        dest="static_folder",
        help="Static folder",
        default=f"{os.getcwd()}ui/build",
    )

    opt_parser.add_argument(
        "-l",
        "--log-file",
        dest="log_file",
        help="Log file name",
        default="",
    )

    opt_parser.add_argument(
        "-L",
        "--log-level",
        dest="log_level",
        help="Log level for all loggers ",
        default="",
    )

    opt_parser.add_argument(
        "-el",
        "--enabled-loggers",
        dest="enabled_logger_list",
        help=(
            "Which loggers to use, default is to use all loggers"
            " ([exception_logger, hwr_logger, mx3_hwr_logger,"
            " user_logger, queue_logger])"
        ),
        default=[
            "exception_logger",
            "hwr_logger",
            "mx3_hwr_logger",
            "user_logger",
            "queue_logger",
            "mx3_ui_logger",
        ],
    )

    opt_parser.add_argument(
        "-w",
        "--ra",
        action="store_true",
        dest="allow_remote",
        help="Enable remote access",
        default=False,
    )

    opt_parser.add_argument(
        "-t",
        "--ra-timeout",
        action="store_true",
        dest="ra_timeout",
        help="Timeout gives control",
        default=False,
    )

    # If `argv` is `None`, then `argparse.ArgumentParser.parse_args`
    # will know to read from `sys.argv` instead.
    return opt_parser.parse_args(argv)


def build_server_and_config(test=False, argv=None):
    cmdline_options = parse_args(argv)

    try:
        # This refactoring (with other bits) allows you to pass a 'path1:path2' lookup path
        # as the hwr_directory. I need it for sensible managing of a multi-beamline test set-up
        # without continuously editing the main config files.
        # Note that the machinery was all there in the core already. rhfogh.
        HWR.init_hardware_repository(cmdline_options.hwr_directory)
        config_path = HWR.get_hardware_repository().find_in_repository("mxcube-web")

        cfg = Config(config_path)

        if test:
            cfg.flask.USER_DB_PATH = "/tmp/mxcube-test-user.db"

        server.init(cmdline_options, cfg)
        mxcube.init(
            server,
            cmdline_options.allow_remote,
            cmdline_options.ra_timeout,
            cmdline_options.log_file,
            cmdline_options.log_level,
            cmdline_options.enabled_logger_list,
            cfg,
        )

        server.register_routes(mxcube)
    except Exception:
        traceback.print_exc()
        raise

    return server, cfg


def main():
    server, cfg = build_server_and_config()
    if server and cfg:
        server.run(cfg)
        return 0
    return 1


if __name__ == "__main__":
    sys.exit(main())



import json

import spectree
from flask import (
    Blueprint,
    Response,
    jsonify,
    request,
)
from mxcubecore import HardwareRepository as HWR

from mxcubeweb.core.models.generic import SimpleNameValue


# Disabling C901 function is too complex (19)
def init_route(app, server, url_prefix):  # noqa: C901
    bp = Blueprint("queue", __name__, url_prefix=url_prefix)

    @bp.route("/start", methods=["PUT"])
    @server.require_control
    @server.restrict
    def queue_start():
        """
        Start execution of the queue.

        :returns: Respons object, status code set to:
                200: On success
                409: Queue could not be started
        """
        sid = request.get_json().get("sid", None)
        app.queue.queue_start(sid)

        return Response(status=200)

    @bp.route("/stop", methods=["PUT"])
    @server.require_control
    @server.restrict
    def queue_stop():
        """
        Stop execution of the queue.

        :returns: Response object status code set to:
                200: On success
                409: Queue could not be stopped
        """
        app.queue.queue_stop()
        return Response(status=200)

    @bp.route("/abort", methods=["PUT"])
    @server.require_control
    @server.restrict
    def queue_abort():
        """
        Abort execution of the queue.

        :returns: Response object, status code set to:
                200 On success
                409 queue could not be aborted
        """
        HWR.beamline.queue_manager.stop()
        return Response(status=200)

    @bp.route("/pause", methods=["PUT"])
    @server.require_control
    @server.restrict
    def queue_pause():
        """
        Pause the execution of the queue

        :returns: Response object, status code set to:
                200: On success
                409: Queue could not be paused
        """
        msg = app.queue.queue_pause()
        server.emit("queue", msg, namespace="/hwr")
        return Response(status=200)

    @bp.route("/unpause", methods=["PUT"])
    @server.require_control
    @server.restrict
    def queue_unpause():
        """
        Unpause execution of the queue

        :returns: Response object, status code set to:
                200: On success
                409: Queue could not be unpause
        """
        msg = app.queue.queue_unpause()
        server.emit("queue", msg, namespace="/hwr")
        return Response(status=200)

    @bp.route("/clear", methods=["PUT", "GET"])
    @server.require_control
    @server.restrict
    def queue_clear():
        """
        Clear the queue.

        :returns: Response object, status code set to:
                200: On success
                409: Queue could not be started
        """
        app.queue.queue_clear()
        return Response(status=200)

    @bp.route("/", methods=["GET"])
    @server.restrict
    def queue_get():
        """
        Get the queue
        :returns: Response object response Content-Type: application/json, json
                object containing the queue on the format returned by
                queue_to_dict. The status code is set to:

                200: On success
                409: On error, could not retrieve queue
        """
        resp = jsonify(app.queue.queue_to_dict(include_lims_data=True))
        resp.status_code = 200
        return resp

    @bp.route("/queue_state", methods=["GET"])
    @server.restrict
    def queue_get_state():
        """
        Get the queue.

        :returns: Response object response Content-Type: application/json, json
                object containing the queue state. The status code is set to:

                200: On success
                409: On error, could not retrieve queue
        """
        resp = jsonify(app.queue.get_queue_state())
        resp.status_code = 200
        return resp

    @bp.route("/<sid>/<tindex>/execute", methods=["PUT"])
    @server.require_control
    @server.restrict
    def execute_entry_with_id(sid, tindex):
        """
        Execute the entry at position (sampleID, task index) in queue
        :param str sid: sampleID
        :param int tindex: task index of task within sample with id sampleID

        :statuscode: 200, no error
                    409, queue entry could not be executed
        """
        try:
            app.queue.execute_entry_with_id(sid, tindex)
        except Exception:
            return Response(status=409)
        else:
            return Response(status=200)

    @bp.route("/", methods=["PUT"])
    @server.require_control
    @server.restrict
    def set_queue():
        app.queue.set_queue(request.get_json())
        return Response(status=200)

    @bp.route("/", methods=["POST"])
    @server.require_control
    @server.restrict
    def queue_add_item():
        tasks = request.get_json()

        queue = app.queue.queue_add_item(tasks)
        sample_list = app.lims.sample_list_get(current_queue=queue)

        resp = jsonify(
            {
                "sampleOrder": queue.get("sample_order", []),
                "sampleList": sample_list.get("sampleList", {}),
            }
        )
        resp.status_code = 200

        server.emit(
            "queue", {"Signal": "update", "message": "observers"}, namespace="/hwr"
        )

        return resp

    @bp.route("/<sqid>/<tqid>", methods=["POST"])
    @server.require_control
    @server.restrict
    def queue_update_item(sqid, tqid):
        data = request.get_json()

        model = app.queue.queue_update_item(sqid, tqid, data)

        resp = jsonify(app.queue.queue_to_dict([model]))
        resp.status_code = 200

        server.emit(
            "queue", {"Signal": "update", "message": "observers"}, namespace="/hwr"
        )

        return resp

    @bp.route("/delete", methods=["POST"])
    @server.require_control
    @server.restrict
    def queue_delete_item():
        item_pos_list = request.get_json()

        app.queue.delete_entry_at(item_pos_list)
        server.emit(
            "queue", {"Signal": "update", "message": "observers"}, namespace="/hwr"
        )

        return Response(status=200)

    @bp.route("/set_enabled", methods=["POST"])
    @server.require_control
    @server.restrict
    def queue_enable_item():
        params = request.get_json()
        qid_list = params.get("qidList", None)
        enabled = params.get("enabled", False)
        app.queue.queue_enable_item(qid_list, enabled)
        server.emit(
            "queue", {"Signal": "update", "message": "observers"}, namespace="/hwr"
        )

        return Response(status=200)

    @bp.route("/<sid>/<ti1>/<ti2>/swap", methods=["POST"])
    @server.require_control
    @server.restrict
    def queue_swap_task_item(sid, ti1, ti2):
        app.queue.swap_task_entry(sid, int(ti1), int(ti2))
        server.emit(
            "queue", {"Signal": "update", "message": "observers"}, namespace="/hwr"
        )

        return Response(status=200)

    @bp.route("/<sid>/<ti1>/<ti2>/move", methods=["POST"])
    @server.require_control
    def queue_move_task_item(sid, ti1, ti2):
        app.queue.move_task_entry(sid, int(ti1), int(ti2))
        server.emit(
            "queue", {"Signal": "update", "message": "observers"}, namespace="/hwr"
        )

        return Response(status=200)

    @bp.route("/sample-order", methods=["POST"])
    @server.require_control
    @server.restrict
    def queue_set_sample_order():
        sample_order = request.get_json().get("sampleOrder", [])
        app.queue.set_sample_order(sample_order)
        server.emit(
            "queue", {"Signal": "update", "message": "observers"}, namespace="/hwr"
        )

        return Response(status=200)

    @bp.route("/<sample_id>", methods=["PUT"])
    @server.require_control
    @server.restrict
    def update_sample(sample_id):
        """
        Update a sample info
            :parameter node_id: entry identifier, integer. It can be a sample
                or a task within a sample
            :request Content-Type: application/json, object containing the
                parameter(s) to be updated, any parameter not sent will
                not be modified.
            :statuscode: 200: no error
            :statuscode: 409: sample info could not be updated, possibly because
                the given sample does not exist in the queue
        """
        params = json.loads(request.data)
        node_id = int(sample_id)

        try:
            app.queue.update_sample(node_id, params)
            resp = jsonify({"QueueId": node_id})
            resp.status_code = 200
            return resp
        except Exception:
            return Response(status=409)

    @bp.route("/<node_id>/toggle", methods=["PUT"])
    @server.require_control
    @server.restrict
    def toggle_node(node_id):
        """
        Toggle a sample or a method checked status
            :parameter id: node identifier, integer
            :statuscode: 200: no error
            :statuscode: 409: node could not be toggled
        """
        app.queue.toggle_node(int(node_id))
        return Response(status=200)

    @bp.route("/available_tasks", methods=["GET"])
    @server.restrict
    def get_avilable_tasks():
        """
        Returns a list of all available tasks
        """
        resp = jsonify(app.queue.get_available_tasks())

        resp.status_code = 200
        return resp

    @bp.route("/update_dependent_field", methods=["POST"])
    @server.restrict
    def update_dependent_field():
        """
        Updates the dependent fields of the given task
        """
        args = request.get_json()
        resp = jsonify(
            app.queue.update_dependent_field(args["task_name"], args["field_data"])
        )

        resp.status_code = 200
        return resp

    @bp.route("/automount", methods=["POST"])
    @server.require_control
    @server.restrict
    def set_autmount():
        data = request.get_json()
        app.queue.set_auto_mount_sample(data.get("automount", False))
        resp = jsonify(data)
        resp.status_code = 200
        return resp

    @bp.route("/num_snapshots", methods=["PUT"])
    @server.require_control
    @server.restrict
    def set_num_snapshots():
        data = request.get_json()
        app.queue.set_num_snapshots(data.get("numSnapshots", app.DEFAULT_NUM_SNAPSHOTS))
        resp = jsonify(
            {"numSnapshots": data.get("numSnapshots", app.DEFAULT_NUM_SNAPSHOTS)}
        )
        resp.status_code = 200

        return resp

    @bp.route("/group_folder", methods=["POST"])
    @server.require_control
    @server.restrict
    def set_group_folder():
        path = request.get_json().get("path", "")

        resp = jsonify(app.queue.set_group_folder(path))
        resp.status_code = 200

        return resp

    @bp.route("/group_folder", methods=["GET"])
    @server.restrict
    def get_group_folder():
        resp = jsonify({"path": HWR.beamline.session.get_group_name()})
        resp.status_code = 200

        return resp

    @bp.route("/auto_add_diffplan", methods=["POST"])
    @server.require_control
    @server.restrict
    def set_autoadd():
        data = request.get_json()
        app.queue.set_auto_add_diffplan(data.get("autoadddiffplan", False))
        resp = jsonify(data)
        resp.status_code = 200
        return resp

    @bp.route("/setting", methods=["POST"])
    @server.require_control
    @server.restrict
    @server.validate(
        json=SimpleNameValue,
        resp=spectree.Response("HTTP_409", "HTTP_200"),
    )
    def set_setting():
        result = app.queue.set_setting(SimpleNameValue(**request.json))

        return jsonify({result[0]: result[1]}) if result else Response(status=409)

    return bp



from mxcubecore.BaseHardwareObjects import HardwareObjectState

from mxcubeweb.core.adapter.adapter_base import ActuatorAdapterBase
from mxcubeweb.core.models.adaptermodels import HOMachineInfoModel
from mxcubeweb.core.util.networkutils import RateLimited


class MachineInfoAdapter(ActuatorAdapterBase):
    """Adapter for MachineInfo like objects"""

    def __init__(self, ho, *args):
        """
        Args:
            (object): Hardware object.
        """
        super().__init__(ho, *args)
        ho.connect("valueChanged", self._value_change)
        self._unique = True

    def _set_value(self, value=None):
        pass

    @RateLimited(0.1)
    def _value_change(self, *args, **kwargs):
        self.value_change(self.get_value(), **kwargs)

    def _get_value(self) -> HOMachineInfoModel:
        return HOMachineInfoModel(value=self.get_attributes())

    def get_attributes(self):
        """Read the information from the HO. Format the output."""
        value_dict = {"current": "-1 mA"}
        try:
            value_dict.update(self._ho.get_value())
            curr = value_dict.get("current")
            if isinstance(curr, float):
                value_dict.update({"current": f"{curr:3.2f} mA"})
        except (TypeError, AttributeError):
            pass
        return value_dict

    def limits(self):
        """
        Returns: The detector distance limits.
        """
        return (-1, -1)

    def stop(self):
        pass

    def state(self):
        return HardwareObjectState.READY.value

    def data(self) -> HOMachineInfoModel:
        return HOMachineInfoModel(**self._dict_repr())



def export(func):
    func._export = True
    func._export_name = func.__name__

    return func


def get_adapter_cls_from_hardware_object(ho):
    # This needs to be a direct import of DataPublisher otherwise the
    # is instance check below fails due to different "import paths" It
    # inly works because mxcubecore adds mxcubecore.HardwareObjects to
    # sys path in __init__.py
    import DataPublisher
    from mxcubecore.HardwareObjects import (
        GenericDiffractometer,
        MiniDiff,
    )
    from mxcubecore.HardwareObjects.abstract import (
        AbstractActuator,
        AbstractBeam,
        AbstractDetector,
        AbstractEnergy,
        AbstractMachineInfo,
        AbstractMotor,
        AbstractNState,
        AbstractShutter,
    )

    from mxcubeweb.core.adapter.actuator_adapter import ActuatorAdapter
    from mxcubeweb.core.adapter.beam_adapter import BeamAdapter
    from mxcubeweb.core.adapter.data_publisher_adapter import DataPublisherAdapter
    from mxcubeweb.core.adapter.detector_adapter import DetectorAdapter
    from mxcubeweb.core.adapter.diffractometer_adapter import DiffractometerAdapter
    from mxcubeweb.core.adapter.energy_adapter import EnergyAdapter
    from mxcubeweb.core.adapter.machine_info_adapter import MachineInfoAdapter
    from mxcubeweb.core.adapter.motor_adapter import MotorAdapter
    from mxcubeweb.core.adapter.nstate_adapter import NStateAdapter

    if isinstance(ho, AbstractNState.AbstractNState | AbstractShutter.AbstractShutter):
        return NStateAdapter
    if isinstance(ho, MiniDiff.MiniDiff | GenericDiffractometer.GenericDiffractometer):
        return DiffractometerAdapter
    if isinstance(ho, AbstractEnergy.AbstractEnergy):
        return EnergyAdapter
    if isinstance(ho, AbstractDetector.AbstractDetector):
        return DetectorAdapter
    if isinstance(ho, AbstractMachineInfo.AbstractMachineInfo):
        return MachineInfoAdapter
    if isinstance(ho, AbstractBeam.AbstractBeam):
        return BeamAdapter
    if isinstance(ho, DataPublisher.DataPublisher):
        return DataPublisherAdapter
    if isinstance(ho, AbstractMotor.AbstractMotor):
        return MotorAdapter
    if isinstance(ho, AbstractActuator.AbstractActuator):
        return ActuatorAdapter
    return None
